# Checklists for various maintenance tasks

## Creating a release

* Update [CHANGELOG.md](CHANGELOG.md) with the major changes since the last
  release
* Update the helm chart template
  * In [Chart.yaml](helm/volsync/Chart.yaml), update `version`, `appVersion`,
    and `annotations.artifacthub.io/changes`
  * Also update the
    [annotation](https://artifacthub.io/docs/topics/annotations/helm/) for the
    signing key if necessary (`annotations.artifacthub.io/signKey`)
* Create a PR w/ the above and commit to `main`
* Branch to `release-X.Y`
* Tag the release: `vX.Y.Z`
* Ensure tagged container images become available on Quay for all containers

### Release the updated Helm chart

* Package the chart:  
  `$ helm package helm/volsync`
* Sign the chart (uses the [GPG Helm
  plugin](https://artifacthub.io/packages/helm-plugin/gpg/gpg)):  
  `$ helm gpg sign volsync-X.Y.Z.tgz`
* Add these files to the `backube/helm-charts` repo:  
  `cd ../helm-charts`  
  `./build-index ../volsync/volsync-X.Y.Z.tgz`
* Create a PR against that repo w/ the changes

## Creating/Updating the operator bundle

* Note: For packaging as an operator, the CSV file is normally auto-generated,
  but some updates may be necessary by hand.  If making updates by hand, edit
  the CSV template in
  [volsync.clusterserviceversion.yaml](config/manifests/bases/volsync.clusterserviceversion.yaml)

* If updates are made to the CSV, any of the CRDs or RBAC (any changes in
  config essentially) then run the bundle command to ensure the bundle manifests
  are up-to-date.

  `$ make bundle`

## Making a new bundle version

If developing a new version, update the [version.mk](version.mk) file and set
the proper values for:

* `VERSION`  
* `CHANNELS` - This should be the set of channels this version should be
  published to  
* `DEFAULT_CHANNEL`  
* `MIN_KUBE_VERSION`

Then run `$ make bundle` again to generate the metadata for the new version

**Note**:  A skip range may need to be specified when creating a new version.

Generally OLM will handle semver updates when updating .Y (for example
from 0.4.0 to 0.4.1 will update automatically within a channel).

However, if the version is to go into a new channel (say, a new channel
`acm-2.6`) then we will need to add replaces or skipRange to the CSV.  As an
example, say we have the following scenario:

* Channel `acm-2.5` with versions `0.4.0` and `0.4.1`  
* User is subscribed to `acm-2.5` and automatically at the head of this channel
  with version `0.4.1` installed  
* Then we make a version `0.4.2` and put this into new channel `acm-2.6`  
* If a user updates their subscription (or ACM does this for them) to channel
  `acm-2.6` then the operator will not automatically update from `0.4.1` to
  `0.4.2`

To get this upgrade to work the easiest solution atm may be to add a skipRange
to the CSV.

The following annotation can be added to the CSV:

```yaml
olm.skipRange: '>=0.4.0 <0.4.2'
```

This will allow upgrades (even between channels) to `0.4.2` from any version
starting from `0.4.0`.  It will also upgrade directly to `0.4.2` without
installing any versions in-between.

See references:

* [operator-framework: how to update operators](https://github.com/operator-framework/operator-lifecycle-manager/blob/master/doc/design/how-to-update-operators.md)

* [olm architecture: creating an update graph](https://olm.operatorframework.io/docs/concepts/olm-architecture/operator-catalog/creating-an-update-graph/)

### Steps to generate a bundle, bundle image, etc for testing

Most of these steps will most likely be automated downstream, but they can be
run manually to generate a bundle at a specific version as well as a bundle
image, catalog source image, etc.

Generally the metadata generated by these commands should not be committed into
this repo

Note for the following commands `VERSION`, `CHANNELS`, `DEFAULT_CHANNEL` have
defaults set in the Makefile so if those values are correct, no need to specify
them

* Generate a bundle at a specific version (0.4.1 in this example) that will be
  published to channel `stable` and channel `acm-2.5`, while making the
  default channel `stable`:

  `$ make bundle VERSION=0.4.1 CHANNELS=stable,acm-2.5
  DEFAULT_CHANNEL=stable`

* To make and push a bundle image (this will be used by an operator catalog)
  that contains the bundle:

  `$ make bundle-build
  BUNDLE_IMG=quay.io/backube/volsync-operator-bundle:v0.4.1`  
  `$ make bundle-push BUNDLE_IMG=quay.io/backube/volsync-operator-bundle:v0.4.1`

* To Make and push a catalog image (this can be done for testing) that will use
  the bundle image in the previous step:

  `$ make catalog-build
  BUNDLE_IMG=quay.io/backube/volsync-operator-bundle:v0.4.1
  CATALOG_IMG=quay.io/backube/test-operator-catalog:v0.4.1`  
  `$ make catalog-push BUNDLE_IMG=quay.io/backube/volsync-operator-bundle:v0.4.1
  CATALOG_IMG=quay.io/backube/test-operator-catalog:v0.4.1`

## Updating tool/dependency versions

* Go
  * Change the version number in [operator.yml](.github/workflows/operator.yml)
  * Change the version number in [go.mod](go.mod)
  * Change the version number for the builder images in
    * [Dockerfile](Dockerfile)
    * [mover-rclone/Dockerfile](mover-rclone/Dockerfile)
    * [mover-restic/Dockerfile](mover-restic/Dockerfile)
  * Update the OpenShift CI builder images & image substitution rules
* [golangci-lint](https://github.com/golangci/golangci-lint/releases)
  * Change the version number in [Makefile](Makefile)
  * Check for added/deprecated linters and adjust [.golangci.yml](.golangci.yml)
    as necessary
* [Helm](https://github.com/helm/helm/releases)
  * Change the version number in [Makefile](Makefile)
* [Kind](https://github.com/kubernetes-sigs/kind/releases)
  * Kind version: Change the version number in
    [operator.yml](.github/workflows/operator.yml)
  * Kubernetes version used by kind:
    * [Available kubernetes
      versions](https://hub.docker.com/r/kindest/node/tags?page=1&ordering=last_updated)
    * Update the default kube version in the
      [setup-kind-cluster.sh](./hack/setup-kind-cluster.sh) script to be the
      latest image
    * Update the matrix list for CI in
      [operator.yml](.github/workflows/operator.yml)
* [Kuttl](https://github.com/kudobuilder/kuttl/releases)
  * Change the version number in [Makefile](Makefile)
* [operator-sdk](https://github.com/operator-framework/operator-sdk/releases)
  * Change the version number in [Makefile](Makefile)
  * Follow the [upgrade
    guide](https://sdk.operatorframework.io/docs/upgrading-sdk-version/) as
    appropriate
* [Rclone](https://github.com/rclone/rclone/releases)
  * Change the version number in
    [mover-rclone/Dockerfile](mover-rclone/Dockerfile) and update GIT hash to match
* [Restic](https://github.com/restic/restic/releases)
  * Change the version number in
    [mover-restic/Dockerfile](mover-restic/Dockerfile), and update GIT hash to
    match
